/* these functions process data from a SHELX HKLF file.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	February 1999
 *
 *  The source file is part of the twindex program, and is under the GPL.
 *  Program: twindex -- a program for analysis of X-ray diffraction data
 *  from non-merohedrally twinned crystals.
 *  Author:  Paul D. Boyle
 *  Copyright (C) 2000  Paul D. Boyle
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#define _XOPEN_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>

#include "twindex.h"

void ProcessShelxFile( PromptData *data )
{

    char input[LINE_LEN],
         *ptr, 
         *dirax_file_name;

    int  count = 0,
         rand_index;
    size_t len;

    char tmp_h[5], tmp_k[5], tmp_l[5];
    char tmp_fsqd[9], tmp_sigma[9];

    int ntmp_h, ntmp_k, ntmp_l;
    float ftmp_fsqd, ftmp_sigma;

    FILE *hklf_file,
         *dirax_file;

    ShelxHKL *refls = NULL,
             *tmp = NULL;

    BisectingAngles bisect;


    hklf_file = fopen( data->file_name, "r" );
    if( !hklf_file ) {
        perror( data->file_name );
        exit( EXIT_FAILURE );
    }
    
/* open the dirax input file */
    if( NULL != (ptr = strchr( data->file_name, '.')) ) {
        len = ptr - data->file_name;
        dirax_file_name = malloc( len + strlen(DIRAX_SUFFIX) + 1 );
        if( !dirax_file_name ) {
            fprintf( stderr, "%s:%d: malloc() failure. Exiting.\n",
                     __FILE__, __LINE__ );
            exit( EXIT_FAILURE );
        }
        strncpy( dirax_file_name, data->file_name, len );
        strcat( dirax_file_name, DIRAX_SUFFIX );
    }
    else
       dirax_file_name = DEFAULT_DIRAX_FILE_NAME;

   dirax_file = fopen( dirax_file_name, "w" );
   if( !dirax_file ) {
       perror( dirax_file_name );
       exit( EXIT_FAILURE );
   }

/* read in the data first, and allocate memory as we go.  If a reflection
 * doesn't meet needed criteria then don't use it.
 */
    while( NULL != fgets(input, sizeof(input), hklf_file) ) {
      if( NULL != (ptr = strchr(input, '\n') ) ) *ptr = '\0';
      sscanf( input, HKLF4_FORMAT, tmp_h, tmp_k, tmp_l,
                     tmp_fsqd, tmp_sigma );
      tmp_h[4] = '\0';
      tmp_k[4] = '\0';
      tmp_l[4] = '\0';
      tmp_fsqd[8] = '\0';
      tmp_sigma[8] = '\0';
      ntmp_h = (int)strtol(tmp_h, NULL, 10);
      ntmp_k = (int)strtol(tmp_k, NULL, 10);
      ntmp_l = (int)strtol(tmp_l, NULL, 10);
      ftmp_fsqd = (float)strtod( tmp_fsqd, NULL );
      ftmp_sigma = (float)strtod( tmp_sigma, NULL );

      /* after processing input line, first test for SHELX 0 0 0 end flag */   
      if( (0 == ntmp_h) && (0 == ntmp_k) && (0 == ntmp_l) ) break;

      /* check criteria */
      if( (data->inet_min > ftmp_fsqd) || 
          (data->inet_sigma_ratio > ftmp_fsqd / ftmp_sigma) ) continue;
      
      tmp = realloc( refls, (count + 1) * sizeof(ShelxHKL) );
      if( !tmp ) {
          fprintf( stderr, "%s:%d -- realloc() failed after %d reflections.\n", 
                   __FILE__, __LINE__, count + 1 );
          exit( EXIT_FAILURE );
      }
      refls = tmp;
      refls[count].used = 0;
      refls[count].hkl[0] = ntmp_h;
      refls[count].hkl[1] = ntmp_k;
      refls[count].hkl[2] = ntmp_l;
      refls[count].f_squared = ftmp_fsqd;
      refls[count].sigma = ftmp_sigma;
      count++;
    }
    fprintf( stdout, "Number of reflections to be used: %d\n", count + 1 );
    fclose( hklf_file );

    if( count + 1 < data->n_list ) {
        data->n_list = count + 1;
        fprintf( stderr, "Warning: not enough data for desired list length."
                 "\nadjusting list length to %d\n", data->n_list );
    }

   fprintf( dirax_file, "%7.5f\n", data->lambda );
   fflush( dirax_file );
    
/* set up random number generation, and select reflections */
   errno = 0;
   srand( (unsigned int)time((time_t *)NULL) );
   if( errno ) {
       fprintf( stderr, "%s:%d Problem with time() call in seeding random number: %s\n",
                         __FILE__, __LINE__, strerror( errno ) );
       exit( EXIT_FAILURE );
   }
          
       
   while( 0 < data->n_list ) {

      rand_index = rand() / (RAND_MAX / count + 1);

      /* check to see if this reflection been used, if so loop again  
       */
      if( refls[rand_index].used != 0 ) continue;
      
      CalculateScatteringVector( data->orient_composite_cell, &refls[rand_index] );

      SVect2Bisect( data->lambda, &bisect, &refls[rand_index] );

      if( (bisect.theta < data->theta_min) ||
          (bisect.theta > data->theta_max) ) {
          refls[rand_index].used = 1;  /* don't use this reflection again */
          continue;
      }
      else {
#ifdef DEBUG
          fprintf( dirax_file, "%6.2f%9.2f%9.2f%9.2f%4.0f%4.0f%4.0f\t%11.6f%11.6f%11.6f\n",
                               bisect.theta, bisect.phi, bisect.chi,
                               bisect.intensity, refls[rand_index].hkl[0],
                               refls[rand_index].hkl[1],
                               refls[rand_index].hkl[2],
                               refls[rand_index].s_vec.c[0],
                               refls[rand_index].s_vec.c[1],
                               refls[rand_index].s_vec.c[2] );
                               
#else
         fprintf( dirax_file, DIRAX_FORMAT, bisect.theta, bisect.phi,
                              bisect.chi, bisect.intensity );
#endif
         fflush( dirax_file );
      }

     /* do some housekeeping at the end of the loop */
      data->n_list--;
      refls[rand_index].used = 1;
   }
   if( NULL != refls )
               free( refls );

   fclose( dirax_file );
   return;    
}
