/* This function takes a SHELX HKLF 4 file and creates an HKLF 5 file.
 * user supplies the transformation matrix which takes the data indexed
 * on the composite cell and indexes it based on the primary twin 
 * fragment.  If a reflection does not index well to the primary fragment
 * it is assumed to come from the secondary fragment.  The twin law is
 * then used to transform that hkl (in primary fragment space) value to
 * h'k'l' (the indicies in the secondary fragment space).  Overlap peaks
 * are determined by observing whether a given scattering vector fits
 * well with both lattices.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	February 1999
 *
 *  The source file is part of the twindex program, and is under the GPL.
 *  Program: twindex -- a program for analysis of X-ray diffraction data
 *  from non-merohedrally twinned crystals.
 *  Author:  Paul D. Boyle
 *  Copyright (C) 2000  Paul D. Boyle
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _XOPEN_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "twindex.h"
#include "output.h"

#ifdef DEBUG
#define DEBUG_FILENAME  "twindex.debug"
#endif

#define N_INTERVAL  1000
#define INTERVAL_TOKEN "."

void ReIndexCompositeData( PromptData *data )
{
    ShelxHKL    *refl;  /* reflections indexed on composite lattice reside
                         * in the 0th elements, and the reflections based
                         * on the n_comp orientation matrices reside in 
                         * the next n_comp elements of the array
                         */

    DiffVector *diff_vec;
    UnitCell  direct_cell, 
              recip_cell;

    double    direct_metric_tensor[3][3],
              recip_metric_tensor[3][3];

    FILE *inp_hklf,
         *out_hkl, 
         *omit_file;

#ifdef DEBUG
    FILE *debug;
#endif

    size_t len;
    int i, j, n,
        k_count = 0,
        n_diffs,
        n_omit = 0,
        n_refl = 0,
        free_mem_flag = 0;

    unsigned int used = 0;

/* some floats for some statistics from the data sets */
    float  avg_f2_sigma = 0.0,
           f2_sigma = 0.0,
           max_f2_sigma = 0.0;
    

    char    tmp_h[5], tmp_k[5], tmp_l[5];

    char tmp_fsqd[9], tmp_sigma[9];

    double determ, 
           (*inverted)[3][3];

    char input[LINE_LEN],
         *ptr,
         *out_file_name;

/* Dump the matricies we have so far so the user can see them. */

   if( 1 == data->composite_indexing ) {
       fputs( "Composite Cell Orientation Matrix\n", stdout );
       Output3x3Matrix( data->orient_composite_cell, stdout );
       ExtractMetricTensors( data->orient_composite_cell,
                             direct_metric_tensor, recip_metric_tensor );
       ExtractUnitCell( &direct_cell, direct_metric_tensor );
       ExtractUnitCell( &recip_cell, recip_metric_tensor );
#ifdef DEBUG
       fprintf( stdout,"Metric Tensor for Composite Lattice\n"  );
       for( j = 0; j < 3; j++ )
            fprintf( stdout, "%9.4f%9.4f%9.4f\n",
                              direct_metric_tensor[j][0],
                              direct_metric_tensor[j][1],
                              direct_metric_tensor[j][2] );
       fprintf( stdout,"Reciprocal Metric Tensor Composite Lattice\n"  );
       for( j = 0; j < 3; j++ )
            fprintf( stdout, "%10.6f%10.6f%10.6f\n",
                              recip_metric_tensor[j][0],
                              recip_metric_tensor[j][1],
                              recip_metric_tensor[j][2] );

#endif
       fprintf( stdout, "Direct Composite Cell:\na = %8.3f b = %8.3f c = %8.3f\n"
                        "alpha = %7.2f beta = %7.2f gamma = %7.2f Vol = %9.2f\n",
                         direct_cell.a, direct_cell.b, direct_cell.c,
                         direct_cell.alpha, direct_cell.beta, direct_cell.gamma,
                         direct_cell.volume );

       fprintf( stdout, "Reciprocal Composite Cell:\na* = %8.6f b* = %8.6f c* = %8.6f\n"
                        "alpha* = %7.2f beta* = %7.2f gamma* = %7.2f Vol = %10.8f\n",
                         recip_cell.a, recip_cell.b, recip_cell.c,
                         recip_cell.alpha, recip_cell.beta, recip_cell.gamma,
                         recip_cell.volume );

       fputs( "\n", stdout );
   }

   for( i = 0; i < data->n_comp; i++ ) {
       fprintf( stdout, "Orientation Matrix for Fragment %d\n", i+1 );
       Output3x3Matrix( data->orient_frag[i], stdout );
       ExtractMetricTensors( data->orient_frag[i],
                             direct_metric_tensor, recip_metric_tensor );
       ExtractUnitCell( &direct_cell, direct_metric_tensor );
       ExtractUnitCell( &recip_cell, recip_metric_tensor );

#ifdef DEBUG
       fprintf( stdout,"Metric Tensor for Fragment %d\n", i+1 );
       for( j = 0; j < 3; j++ )
            fprintf( stdout, "%9.4f%9.4f%9.4f\n",
                              direct_metric_tensor[j][0],
                              direct_metric_tensor[j][1],
                              direct_metric_tensor[j][2] );
       fprintf( stdout,"Reciprocal Metric Tensor for Fragment %d\n", i+1 );
       for( j = 0; j < 3; j++ )
            fprintf( stdout, "%10.6f%10.6f%10.6f\n",
                              recip_metric_tensor[j][0],
                              recip_metric_tensor[j][1],
                              recip_metric_tensor[j][2] );

#endif
       fprintf( stdout, "Direct Cell:\na = %8.3f b = %8.3f c = %8.3f\n"
                        "alpha = %7.2f beta = %7.2f gamma = %7.2f Vol = %9.2f\n",
                         direct_cell.a, direct_cell.b, direct_cell.c,
                         direct_cell.alpha, direct_cell.beta, direct_cell.gamma,
                         direct_cell.volume );

       fprintf( stdout, "Recip Cell:\na* = %8.6f b* = %8.6f c* = %8.6f\n"
                        "alpha* = %7.2f beta* = %7.2f gamma* = %7.2f Vol = %10.8f\n",
                         recip_cell.a, recip_cell.b, recip_cell.c,
                         recip_cell.alpha, recip_cell.beta, recip_cell.gamma,
                         recip_cell.volume );


       fputs( "\n", stdout );
   }

   fputs( "Twin Law\n", stdout );
   Output3x3Matrix( data->twin_law, stdout );


/* calculate the number of elements needed for diff_vec array and
 * allocate the memory for the various data structures.
 */

   n = data->n_comp + 1;
   n_diffs = (n * n - n) / 2;
   diff_vec = malloc( n_diffs * sizeof(DiffVector) );
   if( !diff_vec ) {
       fprintf( stderr, "%s:%d -- malloc() error. Exiting.\n",
                         __FILE__, __LINE__ );
       exit( EXIT_FAILURE );
   }

   inverted = malloc( data->n_comp * sizeof *inverted );
   if( !inverted ) {
       fprintf( stderr, "%s:%d -- malloc() error. Exiting.\n",
                         __FILE__, __LINE__ );
       exit( EXIT_FAILURE );
   }

   refl = malloc( (1 + data->n_comp) * sizeof *refl );
   if( !refl ) {
       fprintf( stderr, "%s:%d -- malloc() error. Exiting.\n",
                         __FILE__, __LINE__ );
       exit( EXIT_FAILURE );
   }

/* open the input file stream */
    inp_hklf = fopen( data->file_name, "r" );
    if( !inp_hklf ) {
        perror( data->file_name );
        exit( EXIT_FAILURE );
    }

/* build HKLF 5 file name and open the stream */
    if( NULL != (ptr = strchr( data->file_name, '.')) ) {
        free_mem_flag = 1;
        len = ptr - data->file_name;
        out_file_name = malloc( len + strlen(HKLF5_SUFFIX) + 1 );
        if( !out_file_name ) {
            fprintf( stderr, "%s:%d: malloc() failure. Exiting.\n",
                     __FILE__, __LINE__ );
            exit( EXIT_FAILURE );
        }
        strncpy( out_file_name, data->file_name, len );
        strcat( out_file_name, HKLF5_SUFFIX );
    }
    else
       out_file_name = DEFAULT_HKL5_FILE_NAME;

    out_hkl = fopen( out_file_name, "w" );
    if( !out_hkl ) {
        perror( out_file_name );
        exit( EXIT_FAILURE );
    }

/* open the OMIT.HKL file */
    omit_file = fopen( OMIT_FILE_NAME, "w" );
    if( !omit_file ) {
        perror( OMIT_FILE_NAME );
        exit( EXIT_FAILURE );
    }
    
#ifdef DEBUG
    debug = fopen( DEBUG_FILENAME, "w" );
    if( !debug ) {
        perror( DEBUG_FILENAME );
        exit( EXIT_FAILURE );
    }
#endif 

/* take the fragments orientation matrices and invert them so
 * that we calculate the Miller indicies for the reflections in terms of
 * this these single lattice spaces.
 */

#ifdef DEBUG 
   fputs( "\n", stdout );
#endif
   for( i = 0; i < data->n_comp; i++ ) {
      determ = Determinant( data->orient_frag[i] );
      InvertMatrix( determ, data->orient_frag[i], inverted[i] );
#ifdef DEBUG
      fprintf( stdout, "Inverted Orientation Matrix for Fragment %d\n", i+1 );
      Output3x3Matrix( inverted[i], stdout );
      fprintf( stdout, "Determinant for Fragment %d Orientation Matrix = %10.6f\n", 
                       i+1, determ );
      fputs( "\n", stdout );
#endif
                           
   }

    fputs( "Reindexing the data", stdout );
    fflush( stdout );
/* process the data file */
    while( NULL != fgets(input, sizeof(input), inp_hklf) ) {
      if( NULL != (ptr = strchr(input, '\n') ) ) *ptr = '\0';

      sscanf( input, HKLF4_FORMAT, tmp_h, tmp_k, tmp_l,
                     tmp_fsqd, tmp_sigma );
      tmp_h[4] = '\0';
      tmp_k[4] = '\0';
      tmp_l[4] = '\0';
      tmp_fsqd[8] = '\0';
      tmp_sigma[8] = '\0';

      /* now fill the composite ShelxHKL struct */
      refl[0].used = 0;  
      refl[0].i_comp = 0;  
      refl[0].hkl[0] = strtod(tmp_h, NULL);
      refl[0].hkl[1] = strtod(tmp_k, NULL);
      refl[0].hkl[2] = strtod(tmp_l, NULL);
      refl[0].f_squared = strtod(tmp_fsqd, NULL);
      refl[0].sigma = strtod(tmp_sigma, NULL);

      /* after processing input line, first test for SHELX 0 0 0 end flag */   
      if( (0 == refl[0].hkl[0]) &&
          (0 == refl[0].hkl[1]) &&
          (0 == refl[0].hkl[2]) ) break;

/* count the values used, give the user eye candy to indicate something is
 * going on.
 */
      n_refl++;
      if( 0 == n_refl % N_INTERVAL ) {
          fputs( INTERVAL_TOKEN, stdout );
          fflush( stdout );
      }

/* After reading in the hkl values based on indexing the composite cell,
 * convert these hkl values to scattering vectors, and convert them to
 * hkl values based on orientation matrices for the fragments.  Round
 * these generally non integer Miller indicies to integers and use them 
 * to calculate scattering vector components for these integer hkl values
 */
   
      CalculateScatteringVector( data->orient_composite_cell, &refl[0] );

#ifdef DEBUG
      fprintf( debug, "      H       K       L    F^2       c1        c2         c3     |c|\n" );
      fprintf( debug, "%8.3f%8.3f%8.3f%8.2f%10.6f%10.6f%10.6f%7.4f\n",
                       refl[0].hkl[0], refl[0].hkl[1], refl[0].hkl[2],
                       refl[0].f_squared, refl[0].s_vec.c[0],refl[0].s_vec.c[1],
                       refl[0].s_vec.c[2], refl[0].s_vec.magnitude );
#endif

/* now loop over the n_comp fragments */
      for( i = 0; i < data->n_comp; i++ ) {
           k_count = i + 1;
           ReIndexReflection( inverted[i], &refl[k_count], refl[0] );
           for( j = 0; j < 3; j++ )
               refl[k_count].hkl[j] = RoundToNearestInteger( refl[k_count].hkl[j] );

           CalculateScatteringVector( data->orient_frag[i], &refl[k_count] );
           refl[k_count].i_comp = 0;

      }

/* look at the difference vectors (of these scattering vectors) and see
 * if they fall within overlap tolerances.
 */      
      k_count = 0;
      for( i = 0; i < data->n_comp; i++ ) 
         for( j = i+1; j <= data->n_comp; j++ ) {
            DifferenceScatteringVector( &diff_vec[k_count].dvec, refl[i].s_vec,
                                        refl[j].s_vec );
            diff_vec[k_count].frag1 = i;
            diff_vec[k_count].frag2 = j;
#ifdef DEBUG
           fprintf( debug,"[%d] (%hd,%hd) %4.0f%4.0f%4.0f\t\t%4.0f%4.0f%4.0f\t%7.4f\n",
                    k_count, diff_vec[k_count].frag1, diff_vec[k_count].frag2, 
                    refl[i].hkl[0],refl[i].hkl[1], refl[i].hkl[2], refl[j].hkl[0],
                    refl[j].hkl[1],refl[j].hkl[2], diff_vec[k_count].dvec.magnitude );
#endif
            k_count++;
         }

#ifdef DEBUG 
       fputs( "\n\n", debug );
#endif
 
/* now test the overlap tolerances */
      used = 0;
      for( i = 0; i < data->n_comp; i++ ) {
           k_count = diff_vec[i].frag2;
           if( data->tolerance > diff_vec[i].dvec.magnitude ) {
               used |= 1 << k_count;
               refl[k_count].i_comp = k_count;
               refl[0].i_comp = 0;  /* this reflection is used. */
           }
      }

      for( i = data->n_comp; i > 0; i-- ) {
         int mult_overlap = 0;
         for( j = i - 1; j > 0; j-- )
              if( used  & (1 << j) ) mult_overlap = 1;
         if( 1 == mult_overlap )
             refl[i].i_comp = -refl[i].i_comp;
      }
      /* if the reflection was not used at all then write it to the 
       * omit file. else, print the reindexed peaks.
       */
      if( 0 == used ) {
          n_omit++;
          f2_sigma = refl[0].f_squared / refl[0].sigma;
          if( f2_sigma > max_f2_sigma )
                         max_f2_sigma = f2_sigma;
          avg_f2_sigma += f2_sigma;

          fprintf( omit_file, "%4.0f%4.0f%4.0f%8.2f%8.2f%8.2f\n",
                    refl[0].hkl[0],
                    refl[0].hkl[1],
                    refl[0].hkl[2],
                    refl[0].f_squared,
                    refl[0].sigma,
                    f2_sigma );
          continue; /*  go back to top of while() loop */
      }
      else 
         OutputHKLF5File( out_hkl, refl, data->n_comp, data->job_type );

    } /* end while() loop */
    fprintf( out_hkl, "   0   0   0     0.0     0.0\n" );

    avg_f2_sigma /= (float)n_omit;

    fprintf( omit_file, "Number of reflections omitted:\t\t%d\n", n_omit );
    fprintf( omit_file, "Maximum F^2 / sigma(F^2) for omitted reflections:\t%5.2f\n", max_f2_sigma );
    fprintf( omit_file, "Average F^2 / sigma(F^2) for omitted reflections:\t%5.2f\n", avg_f2_sigma );

    fprintf( stdout, "\nNumber of reflections processed:\t%d\n", n_refl );           
    fprintf( stdout, "Number of reflections omitted:\t\t%d\n", n_omit );
    fprintf( stdout, "Maximum F^2 / sigma(F^2) for omitted reflections:\t%5.2f\n", max_f2_sigma );
    fprintf( stdout, "Average F^2 / sigma(F^2) for omitted reflections:\t%5.2f\n", avg_f2_sigma );
           

    if( free_mem_flag )
        free( out_file_name );

    fclose( inp_hklf );
    fclose( out_hkl );
    fclose( omit_file );
#ifdef DEBUG
    fclose( debug );
#endif
     
    return;
}
