/*
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	February 1999
 *
 *  This source file is part of the twindex program, and is under the GPL.
 *  Program: twindex -- a program for analysis of X-ray diffraction data
 *  from non-merohedrally twinned crystals.
 *  Author:  Paul D. Boyle
 *  Copyright (C) 2000  Paul D. Boyle
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#define _XOPEN_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "twindex.h"
#include "dialogs.h"

void SelectDialog( PromptData *data )
{
    int default_n_list = DEFAULT_N_LIST;

    data->theta_min = DEFAULT_THETA_MIN;
    data->theta_max = DEFAULT_THETA_MAX;

    TextDialog( "Enter wavelength in Angstroms: ",
                &(data->lambda), 0, NULL, DOUBLE_VALUE  );
    TextDialog( "\nEnter minimum value of F^2 you want to use: ",
                &(data->inet_min), 0, NULL, FLOAT_VALUE );
    TextDialog( "\nEnter minimum value of sigma(F^2) you want to use: ",
                &(data->inet_sigma_ratio), 0, NULL, FLOAT_VALUE );
    TextDialog( "How many reflections to include in DIRAX input file(%d): ",
                &(data->n_list), 0, &default_n_list, INT_VALUE );

    fputs( "\nEnter the orientation matrix with which this data was collected.\n", stdout );
    Get3x3Matrix( data->orient_composite_cell );

    return;
}

void CalculateOverlapsDialog( PromptData *data )
{
     int i,
         ans;

     double default_tol = DEFAULT_TOLERANCE;

     data->composite_indexing = 0;

     TextDialog( "Enter wavelength in Angstroms: ",
                 &(data->lambda), 0, NULL, DOUBLE_VALUE  );
     TextDialog( "\nEnter tolerance factor for overlaps(%4.2f): ",
                 &(data->tolerance), 0, &default_tol, DOUBLE_VALUE );
     TextDialog( "Enter number of components in twin: ",
                 &(data->n_comp), 0, NULL, INT_VALUE  );

/* allocate room for the n_comp number of orientation matricies */
     data->orient_frag = malloc( data->n_comp * sizeof (double [3][3]) );
     if( !data->orient_frag ) {
        fprintf( stderr, "%s:%d -- malloc() error. Exiting.\n", 
                 __FILE__, __LINE__ );
        exit( EXIT_FAILURE );
     }
     fputs( "\nEnter orientation matrix for the twin domain\n"
             "on which you collected intensity data\n", stdout );
     Get3x3Matrix( data->orient_frag[0] );
     do {

        fprintf( stdout, "You need to enter a twin law which relates"
                " the %d fragments,\nor enter %d more orientation matricies for"
                " each individual fragment.\n\n"
                "Enter (t) or (o) to be prompted for the twin law\nor orientation matricies (t|o): ",
                 data->n_comp, data->n_comp - 1 );
        fflush( stdout );
        ans = GetShortAnswer( NULL );
        if( ('t' == ans) || ('o' == ans) ) break;
        else {
             fputs( "Incorrect answer.  Answer t or o to the question.\n", stderr );
             ans = -1;
        }
    } while( -1 == ans );

    switch( ans ) {
      case 't':
         fputs( "\nEnter twin law, T such that H' = T * H\n", stdout );
         Get3x3Matrix( data->twin_law );
         TransformOrientMatrices( data );
         break;
      case 'o':
         for( i = 1; i < data->n_comp; i++ ) {
              fprintf( stdout, "Enter orientation matrix for fragment %d\n", i + 1 );
              Get3x3Matrix( data->orient_frag[i] );
         }
         DeriveTwinLaw( data );
    } /* end of ans switch() */


    return;
}

void ReIndexDialog( PromptData *data )
{
     int i,
         ans1, 
         ans2;

     double default_tol = DEFAULT_TOLERANCE;

     data->composite_indexing = 0;

     TextDialog( "Enter wavelength in Angstroms: ",
                 &(data->lambda), 0, NULL, DOUBLE_VALUE  );
     TextDialog( "\nEnter tolerance factor for overlaps(%4.2f): ",
                 &(data->tolerance), 0, &default_tol, DOUBLE_VALUE );
     TextDialog( "Enter number of components in twin: ",
                 &(data->n_comp), 0, NULL, INT_VALUE  );

/* allocate room for the n_comp number of orientation matricies */
     data->orient_frag = malloc( data->n_comp * sizeof (double [3][3]) );
     if( !data->orient_frag ) {
        fprintf( stderr, "%s:%d -- malloc() error. Exiting.\n", 
                 __FILE__, __LINE__ );
        exit( EXIT_FAILURE );
     }

     do {
         fputs( "Is the intensity data indexed on a composite cell? (y|n) ",
                stdout );
         fflush( stdout );
         ans1 = GetShortAnswer( NULL );
         if( ('y' == ans1) || ('n' == ans1) ) break;
         else {
             fputs( "Incorrect answer.  Answer y or n to the question.\n", stderr );
             ans1 = -1;
         }
     } while( -1 == ans1 );

     switch( ans1 ) {
       case 'y': 
          data->composite_indexing = 1;
          fputs( "\nEnter orientation matrix for composite lattice.\n", stdout );
          Get3x3Matrix( data->orient_composite_cell );

          fputs( "\nEnter orientation matrix for primary twin domain\n", stdout );
          Get3x3Matrix( data->orient_frag[0] );
          do {

              fprintf( stdout, "You need to enter a twin law which relates"
                      " the %d fragments,\nor enter %d more orientation matricies for"
                      " each individual fragment.\n\n"
                      "Enter (t) or (o) to be prompted for the twin law\nor orientation matricies (t|o): ",
                       data->n_comp, data->n_comp - 1 );
              fflush( stdout );
              ans2 = GetShortAnswer( NULL );
              if( ('t' == ans2) || ('o' == ans2) ) break;
              else {
                   fputs( "Incorrect answer.  Answer t or o to the question.\n", stderr );
                   ans2 = -1;
              }
          } while( -1 == ans2 );

          switch( ans2 ) {
            case 't':
               fputs( "\nEnter twin law, T such that H' = T * H\n", stdout );
               Get3x3Matrix( data->twin_law );
               TransformOrientMatrices( data );
               break;
            case 'o':
               for( i = 1; i < data->n_comp; i++ ) {
                    fprintf( stdout, "Enter orientation matrix for fragment %d\n", i + 1 );
                    Get3x3Matrix( data->orient_frag[i] );
               }
               DeriveTwinLaw( data );
          } /* end of ans2 switch() */
          break;
       case 'n':  
          fputs( "To use this option, you needed to have collected data\n"
                 "using a composite cell.  Exiting.\n", stderr );
          exit( EXIT_SUCCESS );
          break;
     }      /* end of ans1 switch() */
    return;
}

/* TextDialog() -- encapsulate the standard, fprintf()/fputs(), fgets(),
 * sscanf() stuff in this function.  For no default value set 'default'
 * to NULL. The argument 'data_type' specifies the type of data to be 
 * filled in 'value'.  The 'val_size' argument only really needs to 
 * be given when the desired input is a character string.  It can 
 * otherwise be safely set to 0.
 *
 * The argument 'prompt' must be a printf() type format statement.
 */

void TextDialog( const char *prompt, void *value, size_t val_size, 
                 void *default_val, int data_type )
{

    char input[LINE_LEN], 
         *ptr;
    int n_scan = 0;

    if( val_size > LINE_LEN ) {
        fprintf( stderr, "%s:%d -- 'val_size = %d exceeds LINE_LEN (%d) parameter."
                         "  Must recompile.\n", __FILE__, __LINE__,
                         val_size, LINE_LEN );
        exit( EXIT_FAILURE );
    } 

    if( NULL != default_val )
        switch( data_type ) {
          case DOUBLE_VALUE:
             fprintf( stdout, prompt, *(double *)default_val );
             break;
          case FLOAT_VALUE:
             fprintf( stdout, prompt, *(float *)default_val );
             break;
          case INT_VALUE:
             fprintf( stdout, prompt, *(int *)default_val );
             break;
          case STRING_VALUE:
             fprintf( stdout, prompt, (char *)default_val );
             break;
          default:
             fprintf( stderr,
                      "Can't handle %d as value of default_type. Exiting.\n",
                       data_type );
             exit( EXIT_SUCCESS );
             break;
        }
    else 
        fputs( prompt, stdout );

   fflush( stdout );
   if( NULL == fgets(input, sizeof(input), stdin) ) {
       fprintf( stderr, "%s:%d --fgets() returned NULL.  Exiting.\n", 
                        __FILE__, __LINE__ );
       exit( EXIT_FAILURE );
   }
   if( NULL != (ptr = strchr(input, '\n') ) ) *ptr = '\0';

   switch( data_type ) {
     case FLOAT_VALUE:
        if( (0 == strcmp(input, "")) && NULL != default_val ) 
            *(float *)value = *(float *)default_val;
        else
            n_scan = sscanf( input, "%f", (float *)value );
            if( !n_scan ) {
                fprintf( stderr, "%s:%d: Bad input -> %d items scanned: %s\n", 
                                  __FILE__,__LINE__, n_scan, input );
                exit( EXIT_FAILURE );
            }
        break;
     case INT_VALUE:
        if( (0 == strcmp(input, "")) && NULL != default_val ) 
            *(int *)value = *(int *)default_val;
        else
            n_scan = sscanf( input, "%d", (int *)value );
            if( !n_scan ) {
                fprintf( stderr, "%s:%d: Bad input -> %d items scanned: %s\n", 
                                  __FILE__,__LINE__, n_scan, input );
                exit( EXIT_FAILURE );
            }
        break;
     case STRING_VALUE: {
        char *val_ptr;
        char *dest = value;
        if( (0 == strcmp(input, "")) && NULL != default_val ) 
            val_ptr = default_val;
        else
            val_ptr = input;
        memcpy( dest, val_ptr, val_size );
        dest[val_size-1] = '\0';
     }
        break;
     case DOUBLE_VALUE:
        if( (0 == strcmp(input, "")) && NULL != default_val ) 
            *(double *)value = *(double *)default_val;
        else
            n_scan = sscanf( input, "%lf", (double *)value );
            if( !n_scan ) {
                fprintf( stderr, "%s:%d: Bad input -> %d items scanned: %s\n", 
                                  __FILE__,__LINE__, n_scan, input );
                exit( EXIT_FAILURE );
            }
        break;
     default:
        fprintf( stderr,
                 "Hmm, why am I here? (file: %s, line %d). Exiting.\n", 
                          __FILE__, __LINE__ );
        exit( EXIT_SUCCESS );
        break;
   }

   return;
}

/* GetShortAnswer() -- gets a single character answer (as an int) and
 * returns it.  If the question has a default answer, that may be specified
 * as an argument to the function.  If no default value, then enter NULL
 * value.  If there is no default and the user hits a return the function
 * returns a -1
 * 
 * This function is good for yes/no sorts of answers, etc;
 */

int GetShortAnswer( int *default_val )
{
    char input[LINE_LEN], 
         *ptr;

     int ret_val = 0;

    if( NULL == fgets(input, sizeof(input), stdin) ) {
        fprintf( stderr, "%s:%d --fgets() returned NULL.  Exiting.\n", 
                         __FILE__, __LINE__ );
        exit( EXIT_FAILURE );
    }
    if( NULL != (ptr = strchr(input, '\n') ) ) *ptr = '\0';

    if( NULL != default_val ) {
        if( 0 == strcmp(input, "") )
            ret_val = *default_val;
        else
            ret_val = input[0];
    }
    else {
        if( 0 == strcmp(input, "") )
            ret_val = -1;
        else
            ret_val = input[0];
    }

    return ret_val;

}

/* Get3x3Matrix() -- prompt user for and receive input for a 3x3 
 * matrix.
 */

void Get3x3Matrix( double m[][3] )
{

    int i, j;
    char input[LINE_LEN], 
         *ptr;

/* first, initialize the matrix */

    for( i = 0; i < 3; i++ )
         for( j = 0; j < 3; j++ )
              m[i][j] = 0.0;


    for( i = 0; i < 3; i++ ) {
       int index = i + 1;
       fprintf( stdout, "Elements %d1 %d2 %d3: ", index, index, index );
       fflush( stdout );
       if( NULL == fgets(input, sizeof(input), stdin) ) {
           fprintf( stderr, "%s:%d --fgets() returned NULL.  Exiting.\n", 
                            __FILE__, __LINE__ );
           exit( EXIT_FAILURE );
       }
       if( NULL != (ptr = strchr(input, '\n') ) ) *ptr = '\0';
       sscanf( input, "%lf%*[\t ,]%lf%*[\t ,]%lf",
               &(m[i][0]), &(m[i][1]), &(m[i][2]) );

    }
    return;
}
      
