/* twindex is a program used for analysis of diffraction patterns of non
 * merohedral twins.  There are three modes of operations:
 * 
 * Select Mode:  program takes a SHELX HKLF file and selects a specified
 * number of reflections at random and creates a file for input into
 * Duisenberg's DIRAX program.  Each reflection is only selected once per
 * list created.  There are several criteria to evaluate the suitability
 * of reflections.
 * 
 * Reindex Mode:  This mode is used when the data was collected on a
 * composite cell. The user is prompted for the orientation matrix with
 * which the data was collected, the number of twin components, and a
 * twin law (or the orientation matrices for the single domain fragments).
 * The program reads a SHELX HKLF 4 file and outputs a SHELX HKLF 5 file.
 * 
 * Overlap Mode:  This mode is used when the data was collected on a single
 * fragment of the twin and the orientation matrix(cies) (or twin law) is
 * (are) known for the other twin fragment(s).  The program takes a SHELX
 * HKLF 4 file and produces a SHELX HKLF5 file.
 * 
 * The progam has been written so that an arbitray number of twin fragments
 * may be analyzed.  The calculation of overlaps is done using a simple
 * "isotropic overlap model".  Which means if two reflections are within
 * a given radius (user chosen), they are considered to be overlapping.
 * This does not take into account the case of partial overlaps.  To do
 * this one would have to deal with the problem at the data reduction step
 * rather than working with SHELX HKLF files.  The present implementation
 * of working with SHELX HKLF files has the virtue of simplicity and 
 * convienence.
 *
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	February 1999
 *
 *  This source file is part of the twindex program, and is under the GPL. 
 *  Program: twindex -- a program for analysis of X-ray diffraction data
 *  from non-merohedrally twinned crystals.
 *  Author:  Paul D. Boyle
 *  Copyright (C) 2000  Paul D. Boyle
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#define _XOPEN_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>   /* for getopt(3) */
#include <errno.h>

/* local headers */
#include "twindex.h"
#include "dialogs.h"

int main( int argc, char **argv )
{
    int i_opt;
    size_t len;

    PromptData prompted;

    prompted.file_name = NULL;
    prompted.dialog = NULL;
    prompted.action = NULL;

    if( argc < 3 ) {
        Usage( argv[0] );
        exit( EXIT_SUCCESS );
    }

    while( (i_opt = getopt(argc, argv, "orsh")) != EOF ) {
      switch( i_opt ) {
        case 'o':
           prompted.job_type = OVERLAP;
           prompted.dialog = CalculateOverlapsDialog;
           prompted.action = CalculateOverlaps;
           break;
        case 's':       /* Assumes a SHELX HKLF 4 file is available */
           prompted.job_type = SELECT;
           prompted.dialog = SelectDialog;
           prompted.action = ProcessShelxFile;
           break;
        case 'r':  /* for reindexing a data set */
           prompted.job_type = REINDEX;
           prompted.dialog = ReIndexDialog;
           prompted.action = ReIndexCompositeData;
           break;
        case 'h':
           Usage( argv[0] );
           exit( EXIT_SUCCESS );
           break;
       default:
           fprintf( stderr, "Sorry, unrecognized option -- %c\n", i_opt );
           Usage( argv[0] );
           exit( EXIT_SUCCESS );
           break;
      }
    }
 
/* fill the non specific prompted data members */   
    len = strlen( argv[argc-1] );
    prompted.file_name = malloc( len + 1 );
    if( !prompted.file_name ) {
         fprintf(stderr, "malloc() error. Exiting,\n" );
         exit( EXIT_FAILURE );
    }
    strncpy( prompted.file_name, argv[argc-1], len  );
    prompted.file_name[len] = '\0';

    if( NULL != prompted.action ) {
        prompted.dialog( &prompted );
        prompted.action( &prompted );
    }
    else
       fprintf( stderr,
                "Sorry, the requested action is not yet implemented.\n" );

    if( NULL != prompted.file_name )
        free( prompted.file_name );

    exit( EXIT_SUCCESS ); 
}
