/* twindex.h -- contains data structures, prototypes, and defines
 * for twindex program.
 * Written by:
 *	Paul D. Boyle
 *	Department of Chemistry
 *	Box 8204
 *	North Carolina State University
 *	Raleigh, NC 27695-8204
 *
 *	February 1999
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */
#define _XOPEN_SOURCE
#include <stdio.h>

#define LINE_LEN  80
#define SELECT   1
#define REINDEX  2
#define OVERLAP  3

#define DEFAULT_EXE_NAME  "twindex"  /* for the Usage() function */


/* some defines used for the TextDialog() function */
#define DEFAULT_NONE  0
#define FLOAT_VALUE   1
#define INT_VALUE     2
#define STRING_VALUE  3
#define DOUBLE_VALUE  4

/* some hopefully reasonable default values */
#define DEFAULT_N_LIST        25
#define DEFAULT_THETA_MIN    1.0
#define DEFAULT_THETA_MAX   25.0
#define DEFAULT_TOLERANCE   0.01

#define FRAG1_OVERLAP  (1 << 0)
#define FRAG2_OVERLAP  (1 << 1)

#define HKLF4_FORMAT  "%4c%4c%4c%8c%8c"
#define DIRAX_FORMAT  "%6.2f%9.2f%9.2f%9.2f\n"
#define DEFAULT_DIRAX_FILE_NAME  "DIRAX.INP"
#define DEFAULT_HKL5_FILE_NAME   "HKLF5.OUT"
#define OMIT_FILE_NAME           "OMIT.HKL"
#define DIRAX_SUFFIX ".drx"
#define HKLF5_SUFFIX  ".hklf5"

#ifndef PI
#define PI 3.141592653589793
#endif

#define RAD2DEG(rad) ((180.0 / PI) * (rad))
#define DEG2RAD(deg) ((PI / 180.) * (deg))

typedef struct prompt_init PromptData;
typedef struct scat_vect S_Vector;
typedef struct diff_vect DiffVector;
typedef struct bisect_angles BisectingAngles;
typedef struct unit_cell UnitCell;
typedef struct shelx_refl  ShelxHKL;

struct scat_vect {    /* Data structure for scatter vector */
        double magnitude;      
        double c[3];          /* scattering vector components */
        };

struct diff_vect {
        short frag1,
              frag2;
        S_Vector dvec;
        };

struct bisect_angles {  /* data structure for bisecting angles */
        double theta,
               phi,
               chi;
        float  intensity;
        };

struct unit_cell {  /* for unit cell parameters,
                             * either direct or reciprocal. */
         float a, 
               b,
               c,
               alpha,
               beta,
               gamma,
               volume;
         }; 


struct prompt_init {
        int job_type;
        int composite_indexing;   /* flag as 1 if data collected using
                                   * composite cell */
        int n_list;               /* # of reflections used in SELECT mode */
        int n_comp;               /* number of twin components */
        double tolerance;         /* residual when reindexing the peaks */
        double lambda;
        double orient_composite_cell[3][3];
        double twin_law[3][3];             /* takes hkl to h'k'l' */
        double (*orient_frag)[3][3];       /* stores orient matrices for n_comp
                                            * components */
        float inet_min;
        float inet_sigma_ratio;
        float theta_min,
              theta_max;
        char *file_name;
        void (*dialog)(PromptData *input);
        void (*action)(PromptData *);
        };

struct shelx_refl {
        unsigned int used;    /* used for the select function */
        int i_comp;           /* used for reindexing functions */ 
        double hkl[3];
        float f_squared;
        float sigma;
        S_Vector s_vec;
        };
       
/* function prototypes */ 
extern void ProcessShelxFile(PromptData *inp);
extern void CalculateOverlaps(PromptData *inp);
extern void ReIndexCompositeData(PromptData *inp);
extern void CalculateScatteringVector(double mat[][3], ShelxHKL *refl);
extern void DifferenceScatteringVector(S_Vector *diff, S_Vector s1, S_Vector s2);
extern void SVect2Bisect(double lambda, BisectingAngles *angles, ShelxHKL *refl);
extern void ReIndexReflection(double inv[][3], ShelxHKL *p, ShelxHKL q);
extern void GenerateTwinIndicies(double twin_law[][3], ShelxHKL *twin, ShelxHKL pri_frag);
extern void TransformOrientMatrices(PromptData *data);
extern void DeriveTwinLaw(PromptData *data);
extern int RoundToNearestInteger(double num);
extern void MatrixMultiplyVector(double prod[3], double mat[][3], double vec[3]);
extern void MultiplyMatricies(double prod[][3], double a[][3], double b[][3]);
extern double DotProduct(double a[], double b[]);
extern double Determinant(double m[][3]);
extern void InvertMatrix(double det, double m[][3], double inv[][3]);
extern void TransposeMatrix(double m[3][3]);
extern void ExtractMetricTensors( double orient[][3],
                                  double direct[][3],
                                  double recip[][3] );
extern void ExtractUnitCell( UnitCell *cell, double m_tensor[3][3] );
extern void Output3x3Matrix( double [][3], FILE *out );
extern void SelectDialog(PromptData *data);
extern void CalculateOverlapsDialog(PromptData *data);
extern void ReIndexDialog(PromptData *data);
extern void TextDialog(const char *prompt, void *value, size_t val_size,
            void *default_val, int data_type);
extern int GetShortAnswer(int *default_val);
extern void Get3x3Matrix(double m[][3]);
extern void Usage( const char * );
